@@ -1,7 +1,7 @@
 package Data::CSel;
 
-our $DATE = '2016-03-23'; # DATE
-our $VERSION = '0.04'; # VERSION
+our $DATE = '2016-04-01'; # DATE
+our $VERSION = '0.05'; # VERSION
 
 use 5.010001;
 use strict;
@@ -312,9 +312,11 @@ our $RE =
   }x;
 
 sub parse_csel {
+    state $re = qr{\A\s*$RE\s*\z};
+
     local $_ = shift;
     local $^R;
-    eval { m{\A\s*$RE\s*\z}; } and return $_;
+    eval { $_ =~ $re } and return $_;
     die $@ if $@;
     return undef;
 }
@@ -510,16 +512,17 @@ sub _simpsel {
             } elsif ($pc eq 'empty') {
                 @res = grep { my @c = Code::Includable::Tree::NodeMethods::_children_as_list($_); !@c } @res;
             } elsif ($pc eq 'has') {
-                @res = grep { csel($opts, $f->{args}[0], $_) }
-                    _uniq_objects(
-                        grep {defined} map { $_->parent } @res);
+                @res = _uniq_objects(
+                    map { reverse Code::Includable::Tree::NodeMethods::ancestors($_) }
+                        csel($opts, $f->{args}[0], @res)
+                    );
             } elsif ($pc eq 'not') {
                 #say "D: res=(".join(",", map {$_->{id}} @res).")";
-                my @all_matches = map { csel($opts, $f->{args}[0], $_) } @res;
+                my @matches = csel($opts, $f->{args}[0], @res);
                 #say "D: all_matches=(".join(",", map {$_->{id}} @all_matches).")";
-                my %all_matches_refaddrs;
-                for (@all_matches) { $all_matches_refaddrs{refaddr($_)}++ }
-                @res = grep { !$all_matches_refaddrs{refaddr($_)} } @res;
+                my %matches_refaddrs;
+                for (@matches) { $matches_refaddrs{refaddr($_)}++ }
+                @res = grep { !$matches_refaddrs{refaddr($_)} } @res;
             } else {
                 die "Unsupported pseudo-class '$pc'";
             }
@@ -607,7 +610,7 @@ Data::CSel - Select tree node objects using CSS Selector-like syntax
 
 =head1 VERSION
 
-This document describes version 0.04 of Data::CSel (from Perl distribution Data-CSel), released on 2016-03-23.
+This document describes version 0.05 of Data::CSel (from Perl distribution Data-CSel), released on 2016-04-01.
 
 =head1 SYNOPSIS
 
